/* Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
 * property and proprietary rights in and to this material, related
 * documentation and any modifications thereto. Any use, reproduction,
 * disclosure or distribution of this material and related documentation
 * without an express license agreement from NVIDIA CORPORATION or
 * its affiliates is strictly prohibited.
 */

struct list_head {
	int x;
};

struct list_node {
	int x;
};

#include "hws_test.h"
#include "../hws_context.h"

#define LOG_CACHE_LINE_SIZE (6)
#define CACHE_LINE_SIZE (1 << LOG_CACHE_LINE_SIZE)
#define BUF_LOG_SIZE 8
#define BUF_SIZE (1 << BUF_LOG_SIZE)
#define BUF_SIZE_BYTES (BUF_SIZE << LOG_CACHE_LINE_SIZE)
#define ASO_WQE_DSEG_SIZE 0x40
#define NUM_OF_QUEUES 16
#define QUEUE_SIZE 256
#define MLX5_CMD_OP_CREATE_MKEY 0x200
#define MLX5_IFC_ARG_SZ_DATA 64

enum aso_wqe_operand_offset {
        ASO_CSEG_DATA_OFFSET = 0,
        ASO_CSEG_COND_OPER_OFFSET = 6,
        ASO_CSEG_COND_1_OFFSET = 8,
        ASO_CSEG_COND_0_OFFSET = 12,
        ASO_CSEG_COND_1_OPER_OFFSET = 16,
        ASO_CSEG_COND_0_OPER_OFFSET = 20,
        ASO_CSEG_DATA_MASK_MODE_OFFSET = 30,
};

struct mlx5_ifc_mkc_bits {
        uint8_t reserved_at_0[0x1];
        uint8_t free[0x1];
        uint8_t reserved_at_2[0x1];
        uint8_t access_mode_4_2[0x3];
        uint8_t reserved_at_6[0x7];
        uint8_t relaxed_ordering_write[0x1];
        uint8_t reserved_at_e[0x1];
        uint8_t small_fence_on_rdma_read_response[0x1];
        uint8_t umr_en[0x1];
        uint8_t a[0x1];
        uint8_t rw[0x1];
        uint8_t rr[0x1];
        uint8_t lw[0x1];
        uint8_t lr[0x1];
        uint8_t access_mode_1_0[0x2];
        uint8_t reserved_at_18[0x8];
        uint8_t qpn[0x18];
        uint8_t mkey_7_0[0x8];
        uint8_t reserved_at_40[0x20];
        uint8_t length64[0x1];
        uint8_t bsf_en[0x1];
        uint8_t sync_umr[0x1];
        uint8_t reserved_at_63[0x2];
        uint8_t expected_sigerr_count[0x1];
        uint8_t reserved_at_66[0x1];
        uint8_t en_rinval[0x1];
        uint8_t pd[0x18];
        uint8_t start_addr[0x40];
        uint8_t len[0x40];
        uint8_t bsf_octword_size[0x20];
        uint8_t reserved_at_120[0x80];
        uint8_t translations_octword_size[0x20];
        uint8_t reserved_at_1c0[0x19];
        uint8_t relaxed_ordering_read[0x1];
        uint8_t reserved_at_1da[0x1];
        uint8_t log_page_size[0x5];
        uint8_t reserved_at_1e0[0x20];
};

struct mlx5_ifc_create_mkey_out_bits {
        uint8_t status[0x8];
        uint8_t reserved_at_8[0x18];
        uint8_t syndrome[0x20];
        uint8_t reserved_at_40[0x8];
        uint8_t mkey_index[0x18];
        uint8_t reserved_at_60[0x20];
};

struct mlx5_ifc_create_mkey_in_bits {
        uint8_t opcode[0x10];
        uint8_t reserved_at_10[0x10];
        uint8_t reserved_at_20[0x10];
        uint8_t op_mod[0x10];
        uint8_t reserved_at_40[0x20];
        uint8_t pg_access[0x1];
        uint8_t reserved_at_61[0x1f];
        struct mlx5_ifc_mkc_bits memory_key_mkey_entry;
        uint8_t reserved_at_280[0x80];
        uint8_t translations_octword_actual_size[0x20];
        uint8_t mkey_umem_id[0x20];
        uint8_t mkey_umem_offset[0x40];
        uint8_t reserved_at_380[0x500];
        uint8_t klm_pas_mtt[][0x20];
};

struct buf_sp {
        struct mlx5dv_devx_umem *umem;
        struct mlx5dv_hws_devx_obj      *mkey;
        uint8_t *buf;
};

struct aso_cseg_input {
        /*
         * operand_masks:
         * [31:30] data_mask_mode
         * [23:20] condition_0_operand
         * [19:16] condition_1_operand
         * [15:12] condition_0_offset
         * [11:8] condition_1_offset
         * [7:6] condition_operand
         * [2:0] data_offset
         */
        uint32_t operand_masks;
        /* Right operand data (condition_0_operand) */
        uint32_t condition_0_data;
        /* Mask applied to condition_0_data and ASO context data */
        uint32_t condition_0_mask;
        /* Right operand data (condition_1_operand) */
        uint32_t condition_1_data;
        /* Mask applied to condition_1_data and ASO context data */
        uint32_t condition_1_mask;
        /* New ASO context data (at offset data_offset) */
        uint64_t bitwise_data;
        /* Mask defining which bits of the ASO context (at offset data_offset) */
        uint64_t data_mask;
};

struct mlx5dv_hws_devx_obj {
        struct mlx5dv_devx_obj *obj;
        uint32_t id;
};

struct send_wqe_aso_ctrl_seg {
        __be64 va_r;
        __be32 lkey;
        __be32 operand_masks;
        __be32 condition_0_data;
        __be32 condition_0_mask;
        __be32 condition_1_data;
        __be32 condition_1_mask;
        __be64 bitwise_data;
        __be64 data_mask;
};

struct aso_wqe {
        struct send_wqe_aso_ctrl_seg aso_ctrl;
        uint8_t bytewise_data[ASO_WQE_DSEG_SIZE];
};

enum aso_op {
        /* AND Operand between Condition 0 and Condition 1 */
        ASO_OPER_LOGICAL_AND = 0x0,
        /* OR Operand between Condition 0 and Condition 1 */
        ASO_OPER_LOGICAL_OR = 0x1,
};

enum aso_data_mask_mode {
        /* Update up to 64 bit. Allows to choose which bit to update*/
        ASO_BITWISE_64BIT = 0x0,
        /* update up to 64 Byte. Allows to choose which Byte to update */
        ASO_BYTEWISE_64BYTE = 0x1,
        /* same as BYTEWISE_64BYTE, however the mask will be logically
         * ORed with a HW calculated mask
         */
        ASO_CALCULATED_64BYTE = 0x2,
};

enum aso_pre_cond_op {
        ASO_OP_ALWAYS_FALSE = 0x0,
        ASO_OP_ALWAYS_TRUE = 0x1,
        ASO_OP_EQUAL = 0x2,
        ASO_OP_NOT_EQUAL = 0x3,
        ASO_OP_GREATER_OR_EQUAL = 0x4,
        ASO_OP_LESSER_OR_EQUAL = 0x5,
        ASO_OP_LESSER = 0x6,
        ASO_OP_GREATER = 0x7,
        ASO_OP_CYCLIC_GREATER = 0x8,
        ASO_OP_CYCLIC_LESSER = 0x9,
};

struct mlx5_ifc_flow_meter_parameter_bits {
        uint8_t valid[0x1];
        uint8_t bucket_overflow[0x1];
        uint8_t start_color[0x2];
        uint8_t both_buckets_on_green[0x1];
        uint8_t meter_mode[0x2];
        uint8_t reserved_0[0x19];

        uint8_t reserved_1[0x20];

        uint8_t reserved_2[0x3];
        uint8_t cbs_exponent[0x5];
        uint8_t cbs_mantissa[0x8];
        uint8_t reserved_3[0x3];
        uint8_t cir_exponent[0x5];
        uint8_t cir_mantissa[0x8];

        uint8_t reserved_4[0x20];

        uint8_t reserved_5[0x3];
        uint8_t ebs_exponent[0x5];
        uint8_t ebs_mantissa[0x8];
        uint8_t reserved_6[0x3];
        uint8_t eir_exponent[0x5];
        uint8_t eir_mantissa[0x8];

        uint8_t reserved_7[0x20];
        uint8_t reserved_8[0x20];
};

struct mlx5_ifc_traffic_counter {
	uint32_t packets_63_32;
	uint32_t packets_31_0;
	uint32_t octets_63_32;
	uint32_t octets_31_0;
};

int buf_create(struct mlx5dv_hws_context *ctx, struct buf_sp *sp,
	       uint8_t opmod);

void buf_destroy(struct buf_sp *sp);

void aso_fill_condition_cseg(struct aso_cseg_input *aso_ctrl,
                             struct send_wqe_aso_ctrl_seg *aso_cseg);

void aso_fill_send_cseg(uint64_t mr_adrr, uint32_t lkey,
                        struct send_wqe_aso_ctrl_seg *aso_cseg,
                        bool read);

