# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

#cython: language_level=3

from pyverbs.providers.mlx5.hws_resource cimport HwsResource
from pyverbs.providers.mlx5.hws_context cimport HwsContext
from pyverbs.providers.mlx5.hws_table cimport HwsTable
from pyverbs.providers.mlx5.mlx5dv cimport Mlx5DevxObj
from pyverbs.base cimport PyverbsCM, PyverbsObject
cimport pyverbs.providers.mlx5.libmlx5hws as hws


cdef class HwsAction(PyverbsCM):
    cdef hws.mlx5dv_hws_action *action
    cdef HwsContext context
    cdef object rules
    cdef object matchers
    cdef object used_actions
    cdef add_ref(self, obj)
    cdef del_ref(self, obj)


cdef class HwsActionDestDrop(HwsAction):
    pass


cdef class HwsActionDestTir(HwsAction):
    cdef Mlx5DevxObj devx_obj


cdef class HwsActionDestMatcher(HwsAction):
    pass


cdef class HwsActionTag(HwsAction):
    pass


cdef class HwsActionDestRoot(HwsAction):
    pass


cdef class HwsActionDestDefaultMiss(HwsAction):
    pass


cdef class HwsActionDestTable(HwsAction):
    cdef HwsTable table


cdef class HwsActionDestVport(HwsAction):
    pass


cdef class HwsActionCounter(HwsAction):
    cdef list counters


cdef class HwsActionPopVlan(HwsAction):
    pass


cdef class HwsActionPushVlan(HwsAction):
    pass


cdef class HwsActionModify(HwsAction):
    cdef list arguments


cdef class HwsActionReformatTnlL2ToL2(HwsAction):
    pass


cdef class HwsActionReformatL2ToTnlL2(HwsAction):
    cdef list arguments


cdef class HwsActionReformatTnlL3ToL2(HwsAction):
    cdef list arguments


cdef class HwsActionReformatL2ToTnlL3(HwsAction):
    cdef list arguments


cdef class HwsActionInsertHeaderAttr(PyverbsCM):
    cdef hws.mlx5dv_hws_action_insert_header hdr


cdef class HwsActionInsertHeader(HwsAction):
    cdef list arguments


cdef class HwsActionRemoveHeaderAttr(PyverbsObject):
    cdef hws.mlx5dv_hws_action_remove_header_attr attr


cdef class HwsActionRemoveHeader(HwsAction):
    cdef list arguments


cdef class HwsActionAsoMeter(HwsAction):
    cdef list aso_meters


cdef class HwsActionAsoCT(HwsAction):
    cdef list resources


cdef class HwsActionDestArrayAttr(PyverbsCM):
    cdef HwsAction dest_action
    cdef hws.mlx5dv_hws_action_dest_attr attr


cdef class HwsActionDestArray(HwsAction):
    cdef object actions


cdef class HwsActionEnqueueArgWriteAttr(PyverbsCM):
    cdef hws.mlx5dv_hws_action_enqueue_arg_write_attr attr


cdef class HwsActionTemplateAttr(PyverbsObject):
    cdef hws.mlx5dv_hws_at_attr attr


cdef class HwsActionTemplate(PyverbsCM):
    cdef hws.mlx5dv_hws_at *action_template
    cdef object actions


cdef class HwsActionData(PyverbsCM):
    cdef hws.mlx5dv_hws_action_data action_data
    cdef void* ptr_to_free
