# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

from pyverbs.pyverbs_error import PyverbsError
from libc.stdlib cimport free, calloc
from libc.stdint cimport uint8_t
from libc.string cimport memcpy


cdef class HwsData(PyverbsCM):
    def __init__(self, match, mask, last=None):
        """
        Initializes a HwsData object representing mlx5dv_hws_item_data struct.
        :param match: Buffer of match values.
        :param mask: Buffer of masks for match.
        :param last: Buffer of range from match to last with mask.
        """
        cdef char *py_bytes_c
        super().__init__()
        self.data.match = NULL
        self.data.mask = NULL
        self.data.last = NULL
        size = match.size
        # TODO: make it a for loop - Scrum Task #4003179
        if match:
            self.data.match = <void *>calloc(1,size)
            if self.data.match == NULL:
                raise PyverbsError(f'Failed to allocate buffer of size {size}.')
            py_bytes = bytes(match)
            py_bytes_c = <char*> py_bytes
            memcpy(<void *>self.data.match, py_bytes_c, size)
        if mask:
            self.data.mask = <void *>calloc(1,size)
            if self.data.mask == NULL:
                raise PyverbsError(f'Failed to allocate buffer of size {size}.')
            py_bytes = bytes(mask)
            py_bytes_c = <char*> py_bytes
            memcpy(<void *>self.data.mask, py_bytes_c, size)
        if last:
            self.data.last = <void *>calloc(1,size)
            if self.data.last == NULL:
                raise PyverbsError(f'Failed to allocate buffer of size {size}.')
            py_bytes = bytes(last)
            py_bytes_c = <char*> py_bytes
            memcpy(<void *>self.data.last, py_bytes_c, size)

    def __dealloc__(self):
        self.close()

    cpdef close(self):
        if self.logger:
            self.logger.debug('Closing HwsData.')
        if self.data.match != NULL:
            free(<void *>self.data.match)
            self.data.match = NULL
        if self.data.mask != NULL:
            free(<void *>self.data.mask)
            self.data.mask = NULL
        if self.data.last != NULL:
            free(<void *>self.data.last)
            self.data.last = NULL
