# Copyright (c) <2024> NVIDIA CORPORATION & AFFILIATES. All rights reserved.
#
# NVIDIA CORPORATION, its affiliates and licensors retain all intellectual
# property and proprietary rights in and to this material, related
# documentation and any modifications thereto. Any use, reproduction,
# disclosure or distribution of this material and related documentation
# without an express license agreement from NVIDIA CORPORATION or
# its affiliates is strictly prohibited.

#cython: language_level=3

from libc.stdint cimport uint8_t, uint16_t, uint32_t, uint64_t, int32_t
cimport pyverbs.providers.mlx5.mlx5hws_enums as e
cimport pyverbs.libibverbs as v
cimport libc.stdio as s


cdef extern from 'infiniband/mlx5dv.h':
    cdef struct mlx5dv_devx_obj


cdef extern from 'infiniband/mlx5dv_hws.h':

    cdef struct mlx5dv_hws_context
    cdef struct mlx5dv_hws_table
    cdef struct mlx5dv_hws_action
    cdef struct mlx5dv_hws_mt
    cdef struct mlx5dv_hws_at
    cdef struct mlx5dv_hws_matcher
    cdef struct mlx5dv_hws_rule
    cdef struct mlx5dv_hws_resource
    cdef struct mlx5dv_hws_parser_graph
    cdef struct mlx5dv_hws_parser_node
    cdef struct mlx5dv_hws_parser_arc
    cdef struct mlx5dv_hws_parser_sampler

    cdef struct mlx5dv_hws_context_attr:
        uint32_t comp_mask
        uint16_t queues
        uint16_t queue_size
        v.ibv_pd *pd
        v.ibv_context *shared_ibv_ctx

    cdef struct mlx5dv_hws_table_attr:
        uint32_t comp_mask
        e.mlx5dv_hws_table_type type
        uint32_t level
        uint32_t ib_port

    cdef struct mlx5dv_hws_item_data:
        const void *match
        const void *mask
        const void *last

    cdef struct mlx5dv_hws_field_match_type_to_value:
        uint32_t bit_off
        void *user_conv_data
        void (*user_conv_func)(const mlx5dv_hws_item_data *item_data,
                               void *user_conv_data,
                               e.mlx5dv_hws_field_user_conv_flag conv_flag,
                               uint32_t *ret_conv_data)

    cdef struct mlx5dv_hws_field_match_type_to_field:
        e.mlx5dv_hws_field_name fname_b
        e.mlx5dv_hws_field_header header_b

    cdef struct mlx5dv_hws_field:
        e.mlx5dv_hws_field_name fname
        e.mlx5dv_hws_field_header header
        e.mlx5dv_hws_field_match_op op
        e.mlx5dv_hws_field_match_type type
        mlx5dv_hws_field_match_type_to_value value
        mlx5dv_hws_field_match_type_to_field field

    cdef struct mlx5dv_hws_item:
        mlx5dv_hws_item_data data
        uint8_t num_of_fields
        mlx5dv_hws_field *fields

    cdef struct mlx5dv_hws_mt_attr:
        uint32_t comp_mask

    cdef struct table:
        uint8_t sz_row_log
        uint8_t sz_col_log

    cdef struct rule:
        uint8_t num_log

    cdef struct mlx5dv_hws_matcher_attr:
        uint32_t comp_mask
        uint32_t priority
        e.mlx5dv_hws_matcher_insert_mode insert_mode
        e.mlx5dv_hws_matcher_distribute_mode distribute_mode
        e.mlx5dv_hws_matcher_resource_mode resource_mode
        table table
        rule rule
        bint optimize_using_rule_idx
        bint resizable
        bint isolated
        bint always_hit
        uint8_t max_num_of_at_attach

    cdef struct mlx5dv_hws_rule_attr:
        uint32_t comp_mask
        uint16_t queue_id
        void *user_data
        uint32_t rule_idx
        uint32_t burst

    cdef struct mlx5dv_hws_resource_attr:
        uint32_t comp_mask
        e.mlx5dv_hws_resource_type type
        size_t bulk_log_size
        uint8_t block_alloc_support

    ctypedef unsigned int __be32

    cdef struct mlx5dv_hws_resource_enqueue_aso_attr:
        uint32_t comp_mask
        bint burst
        void *user_data
        int32_t offset

    cdef struct mlx5dv_hws_at_attr:
        uint32_t comp_mask

    cdef struct _tag:
        uint32_t value

    cdef struct _push_vlan:
        __be32 vlan_hdr

    cdef struct _modify_header:
        uint32_t resource_offset
        uint8_t resource_idx
        uint64_t *actions_data

    cdef struct _reformat_header:
        uint32_t resource_offset
        uint8_t resource_idx
        uint8_t *hdr_data

    cdef struct _counter:
        uint32_t resource_offset
        uint8_t *aso_wqe
        uint8_t resource_idx

    cdef struct _aso_meter:
        uint32_t resource_offset
        uint8_t resource_idx
        e.mlx5dv_hws_action_aso_meter_color init_color
        uint8_t *aso_wqe

    cdef struct _aso_ct:
        uint32_t resource_offset
        uint8_t resource_idx
        e.mlx5dv_hws_action_aso_ct_direction direction
        uint8_t *aso_wqe

    cdef struct _dest_matcher:
        uint32_t index

    cdef struct mlx5dv_hws_action_data:
        _tag tag
        _push_vlan push_vlan
        _modify_header modify_header
        _reformat_header reformat_l2_to_tnl_l2
        _reformat_header reformat_l2_to_tnl_l3
        _reformat_header reformat_tnl_l3_to_l2
        _reformat_header insert_header
        _counter counter
        _aso_meter aso_meter
        _aso_ct aso_ct
        _dest_matcher dest_matcher

    cdef struct mlx5dv_hws_queue_op_result:
        e.mlx5dv_hws_queue_op_status status
        void *user_data

    cdef struct table_info:
        uint32_t ft_id
        uint32_t ft_fw_type
        uint32_t level

    cdef struct mlx5dv_hws_metric_table:
        table_info info

    cdef struct mlx5dv_hws_parser_field_cfg:
        uint32_t bit_offset
        uint32_t bit_length

    cdef struct _fixed:
        uint32_t hdr_len

    cdef struct _field:
        mlx5dv_hws_parser_field_cfg field
        uint32_t multiplier
        uint32_t addition

    cdef struct mlx5dv_hws_parser_hdr_len_cfg:
        e.mlx5dv_hws_parser_hdr_len_mode mode
        _fixed fixed
        _field field

    cdef struct mlx5dv_hws_parser_node_cfg:
        mlx5dv_hws_parser_hdr_len_cfg hdr_len
        bint has_options
        bint has_selection_field
        uint32_t option_offset
        mlx5dv_hws_parser_hdr_len_cfg option_len
        mlx5dv_hws_parser_field_cfg option_type_field
        mlx5dv_hws_parser_field_cfg selection_field

    cdef struct _native:
        e.mlx5dv_hws_parser_native_node node

    cdef struct _flex:
        mlx5dv_hws_parser_node *node

    cdef struct mlx5dv_hws_parser_arc_node_cfg:
        e.mlx5dv_hws_parser_node_type type
        _native native
        _flex flex

    cdef struct mlx5dv_hws_parser_arc_cfg:
        mlx5dv_hws_parser_arc_node_cfg src_node
        mlx5dv_hws_parser_arc_node_cfg dst_node
        uint32_t src_encap_level
        uint32_t selection_field_data
        bint is_inner_transition

    cdef struct mlx5dv_hws_parser_sampler_cfg:
        mlx5dv_hws_parser_sampler *share_hw_resources
        uint32_t node_encap_level
        bint sample_from_option
        uint32_t option_type
        mlx5dv_hws_parser_field_cfg field_cfg

    cdef struct mlx5dv_hws_parser_geneve_option_cfg:
        uint16_t option_class
        uint8_t option_type
        uint8_t option_data_len
        uint8_t option_class_ignore
        uint8_t offset_valid
        uint8_t sample_offset

    cdef struct mlx5dv_hws_action_modify_header:
        size_t sz
        uint64_t *modify_actions

    cdef struct mlx5dv_hws_action_hdr_data:
        size_t sz
        uint8_t *hdr_data

    cdef struct mlx5dv_hws_action_enqueue_arg_write_attr:
        uint32_t comp_mask
        void *user_data
        size_t resource_offset
        size_t data_size
        uint8_t *arg_data
        uint32_t burst

    cdef struct mlx5dv_hws_action_insert_header:
        mlx5dv_hws_action_hdr_data hdr_data
        uint8_t anchor
        uint8_t offset
        uint8_t encap
        uint8_t push_esp
        uint8_t skip_reparse

    cdef struct _by_header:
        uint8_t start_anchor
        uint8_t end_anchor
        uint8_t decap

    cdef struct _by_offset:
        uint8_t start_anchor
        uint8_t size

    cdef struct mlx5dv_hws_action_remove_header_attr:
        e.mlx5dv_hws_action_remove_header_type type
        _by_header by_header
        _by_offset by_offset

    cdef struct _reformat_action:
        e.mlx5dv_hws_action_dest_reformat_type type
        mlx5dv_hws_action_hdr_data hdr_data

    cdef struct mlx5dv_hws_action_dest_attr:
        mlx5dv_hws_action *dest_action
        _reformat_action reformat

    cdef struct mlx5dv_hws_rule_query_data:
        uint32_t comp_mask
        size_t match_tag_sz
        uint8_t match_tag[64]
    cdef struct matcher_info:
        uint64_t matcher_tables_mem_sz
        uint32_t matcher_id
        uint32_t priority
        uint8_t log_num_of_rows
        uint8_t log_num_of_columns
        bint col_matcher_exist
        uint8_t col_matcher_log_num_of_rows
        uint8_t col_matcher_log_num_of_columns
        uint8_t num_of_match_stes
        uint8_t num_of_action_stes
        bint is_jumbo

    cdef struct matcher_cost:
        uint64_t memory
        uint32_t miss_pkt_processing

    cdef struct mlx5dv_hws_metric_matcher:
        matcher_info info
        matcher_cost cost

    cdef struct mt_resource:
        uint8_t match_stes
        uint8_t action_stes
        uint8_t counters
        uint8_t aso
        uint8_t args
        uint8_t crypto

    cdef struct mt_cost:
        uint64_t memory
        uint32_t pkt_processing
        uint32_t insertion

    cdef struct mlx5dv_hws_metric_matcher_template:
        mt_resource resource
        mt_cost cost

    cdef struct mlx5dv_hws_metric_matcher_template_attr:
        uint64_t flags
        mlx5dv_hws_action *dest_action
        mlx5dv_hws_action_data *actions_data
        uint8_t mt_idx
        uint8_t at_idx

    cdef mlx5dv_hws_context *mlx5dv_hws_context_open(v.ibv_context *ibv_ctx,
                                                     mlx5dv_hws_context_attr *attr)

    cdef int mlx5dv_hws_context_close(mlx5dv_hws_context *ctx)

    cdef mlx5dv_hws_table *mlx5dv_hws_table_create(mlx5dv_hws_context *ctx,
                                                   mlx5dv_hws_table_attr *attr)

    cdef int mlx5dv_hws_table_destroy(mlx5dv_hws_table *tbl)

    cdef int mlx5dv_hws_table_set_default_miss(mlx5dv_hws_table *tbl, mlx5dv_hws_table *miss_tbl)

    cdef mlx5dv_hws_mt *mlx5dv_hws_match_template_create(mlx5dv_hws_context *ctx,
                                                         const mlx5dv_hws_item items[],
                                                         uint8_t num_of_items,
                                                         mlx5dv_hws_mt_attr *attr)

    cdef int mlx5dv_hws_match_template_destroy(mlx5dv_hws_mt *mt)

    cdef mlx5dv_hws_at *mlx5dv_hws_action_template_create(mlx5dv_hws_action** actions,
                                                          size_t num_of_actions,
                                                          mlx5dv_hws_at_attr *attr)

    cdef void mlx5dv_hws_action_template_destroy(mlx5dv_hws_at *at)

    cdef mlx5dv_hws_matcher *mlx5dv_hws_matcher_create(mlx5dv_hws_table *tbl,
                                                       mlx5dv_hws_mt *mt[],
                                                       uint8_t num_of_mt,
                                                       mlx5dv_hws_at *at[],
                                                       uint8_t num_of_at,
                                                       mlx5dv_hws_matcher_attr *attr)

    cdef int mlx5dv_hws_matcher_destroy(mlx5dv_hws_matcher *matcher)

    int mlx5dv_hws_matcher_resize_set_target(mlx5dv_hws_matcher *src_matcher,
                                             mlx5dv_hws_matcher *dst_matcher)

    int mlx5dv_hws_matcher_resize_rule_move(mlx5dv_hws_matcher *src_matcher, mlx5dv_hws_rule *rule,
                                            mlx5dv_hws_rule_attr *attr)

    int mlx5dv_hws_matcher_attach_at(mlx5dv_hws_matcher *matcher, mlx5dv_hws_at *at)

    cdef size_t mlx5dv_hws_rule_get_handle_size()

    cdef int mlx5dv_hws_rule_create(mlx5dv_hws_matcher *matcher, uint8_t mt_idx,
                                    mlx5dv_hws_item_data item_data[], uint8_t at_idx,
                                    mlx5dv_hws_action_data action_data[],
                                    mlx5dv_hws_action *dest_action, mlx5dv_hws_rule_attr *attr,
                                    mlx5dv_hws_rule *rule_handle)

    cdef int mlx5dv_hws_rule_action_data_update(mlx5dv_hws_rule *rule_handle, uint8_t at_idx,
                                                mlx5dv_hws_action_data actions_data[],
                                                mlx5dv_hws_action *dest_action,
                                                mlx5dv_hws_rule_attr *attr)

    cdef int mlx5dv_hws_rule_destroy(mlx5dv_hws_rule *rule, mlx5dv_hws_rule_attr *attr)

    cdef mlx5dv_hws_resource *mlx5dv_hws_resource_alloc(mlx5dv_hws_context *ctx,
                                                        mlx5dv_hws_resource_attr *attr)

    cdef int mlx5dv_hws_resource_free(mlx5dv_hws_resource *resource)

    cdef int mlx5dv_hws_resource_max_block_size(mlx5dv_hws_resource *resource)

    cdef int mlx5dv_hws_resource_get_offset(mlx5dv_hws_resource *resource, uint16_t qid,
                                            uint32_t *resource_offset)

    cdef int mlx5dv_hws_resource_put_offset(mlx5dv_hws_resource *resource, uint16_t qid,
                                            uint32_t resource_offset)

    cdef int mlx5dv_hws_resource_get_offset_block(mlx5dv_hws_resource *resource, uint16_t qid,
                                                  uint8_t log_sz, uint32_t *resource_offset)

    cdef int mlx5dv_hws_resource_put_offset_block(mlx5dv_hws_resource *resource, uint16_t qid,
                                                  uint8_t log_sz, uint32_t resource_offset)

    cdef int mlx5dv_hws_resource_get_devx_obj(mlx5dv_hws_resource *resource,
                                              mlx5dv_devx_obj **ret_obj, uint32_t *ret_obj_id)

    cdef int mlx5dv_hws_resource_enqueue_aso_wqe(mlx5dv_hws_resource *resource, uint16_t qid,
                                                 __be32 *aso_wqe, size_t wqe_len,
                                                 mlx5dv_hws_resource_enqueue_aso_attr *attr)

    cdef int mlx5dv_hws_action_flags_set_ib_port(mlx5dv_hws_context *ctx, uint32_t ib_port,
                                                 uint64_t *action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_dest_drop(mlx5dv_hws_context *ctx,
                                                               uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_dest_table(mlx5dv_hws_context *ctx,
                                                                mlx5dv_hws_table *tbl,
                                                                uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_dest_default_miss(mlx5dv_hws_context *ctx,
                                                                       uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_dest_tir(mlx5dv_hws_context *ctx,
                                                              mlx5dv_devx_obj *obj,
                                                              bint is_local, uint64_t action_flags)

    cdef mlx5dv_hws_action * \
    mlx5dv_hws_action_create_dest_matcher(mlx5dv_hws_context *ctx, mlx5dv_hws_matcher *matcher,
                                          e.mlx5dv_hws_action_dest_matcher_type type,
                                          uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_dest_root(mlx5dv_hws_context *ctx,
                                                               uint16_t priority,
                                                               uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_dest_vport(mlx5dv_hws_context *ctx,
                                                                uint32_t ib_port_num,
                                                                uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_tag(mlx5dv_hws_context *ctx,
                                                         uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_counter(mlx5dv_hws_context *ctx,
                                                             mlx5dv_hws_resource *resource[],
                                                             uint8_t num_of_resources,
                                                             uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_pop_vlan(mlx5dv_hws_context *ctx,
                                                              uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_push_vlan(mlx5dv_hws_context *ctx,
                                                               uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_modify_header(
                    mlx5dv_hws_context *ctx, mlx5dv_hws_resource *resource[],
                    uint8_t num_of_resources, mlx5dv_hws_action_modify_header *pattern,
                    bint force_reparse,
                    uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_reformat_tnl_l2_to_l2(mlx5dv_hws_context *ctx,
                                                                           uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_reformat_l2_to_tnl_l2(
                    mlx5dv_hws_context *ctx, mlx5dv_hws_resource *resource[],
                    uint8_t num_of_resources, mlx5dv_hws_action_hdr_data *hdr_data,
                    uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_reformat_l2_to_tnl_l3(
                    mlx5dv_hws_context *ctx, mlx5dv_hws_resource *resource[],
                    uint8_t num_of_resources, mlx5dv_hws_action_hdr_data *hdr_data,
                    uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_reformat_tnl_l3_to_l2(
                    mlx5dv_hws_context *ctx, mlx5dv_hws_resource *resource[],
                    uint8_t num_of_resources, mlx5dv_hws_action_hdr_data *hdr_data,
                    uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_aso_meter(mlx5dv_hws_context *ctx,
                                                               mlx5dv_hws_resource *resource[],
                                                               uint8_t num_of_resources,
                                                               uint8_t return_reg_id,
                                                               uint64_t action_flags)

    cdef mlx5dv_hws_action *mlx5dv_hws_action_create_aso_ct(mlx5dv_hws_context *ctx,
                                                            mlx5dv_hws_resource *resource[],
                                                            uint8_t num_of_resources,
                                                            uint8_t return_reg_id,
                                                            uint64_t action_flags)

    cdef mlx5dv_hws_action * \
    mlx5dv_hws_action_create_insert_header(mlx5dv_hws_context *ctx,
                                        mlx5dv_hws_resource *resource[],
                                        uint8_t num_of_resources,
                                        mlx5dv_hws_action_insert_header *hdr,
                                        uint64_t action_flags)

    cdef mlx5dv_hws_action * \
    mlx5dv_hws_action_create_remove_header(mlx5dv_hws_context *ctx,
                                        mlx5dv_hws_action_remove_header_attr *attr,
                                        uint64_t action_flags)

    cdef mlx5dv_hws_action * \
    mlx5dv_hws_action_create_dest_array(mlx5dv_hws_context *ctx, size_t num_dest,
                                        mlx5dv_hws_action_dest_attr dests[], uint64_t action_flags)

    cdef void mlx5dv_hws_action_destroy(mlx5dv_hws_action *action)

    cdef int mlx5dv_hws_action_enqueue_arg_write(mlx5dv_hws_action *action,
                                                 mlx5dv_hws_resource *resource, uint16_t qid,
                                                 mlx5dv_hws_action_enqueue_arg_write_attr *attr)

    cdef int mlx5dv_hws_queue_pending(mlx5dv_hws_context *ctx, uint16_t queue_id)

    cdef int mlx5dv_hws_rule_query(mlx5dv_hws_rule *rule, mlx5dv_hws_rule_query_data *ret_query)

    cdef int mlx5dv_hws_rule_hash_calculate(mlx5dv_hws_matcher *matcher,
                                            mlx5dv_hws_item_data item_data[], uint8_t mt_idx,
                                            uint32_t *ret_hash_idx,
                                            uint32_t *ret_hash_idx)

    cdef int mlx5dv_hws_queue_poll(mlx5dv_hws_context *ctx, uint16_t queue_id,
                                   mlx5dv_hws_queue_op_result res[], uint32_t res_nb)

    cdef int mlx5dv_hws_queue_execute_op(mlx5dv_hws_context *ctx, uint16_t queue_id,
                                         uint32_t queue_op)

    cdef int mlx5dv_hws_metric_query_table(mlx5dv_hws_table *tbl, uint64_t flags,
                                           mlx5dv_hws_metric_table *out)

    cdef int mlx5dv_hws_metric_query_matcher(mlx5dv_hws_matcher *matcher, uint64_t flags,
                                             mlx5dv_hws_metric_matcher *out)

    cdef int mlx5dv_hws_metric_query_matcher_template(mlx5dv_hws_matcher *matcher,
                                                      mlx5dv_hws_metric_matcher_template_attr *attr,
                                                      mlx5dv_hws_metric_matcher_template *out)

    cdef int mlx5dv_hws_debug_dump(mlx5dv_hws_context *ctx, s.FILE *f)

    cdef mlx5dv_hws_parser_graph *mlx5dv_hws_parser_graph_create(mlx5dv_hws_context *ctx)

    cdef int mlx5dv_hws_parser_graph_destroy(mlx5dv_hws_parser_graph *graph)

    cdef mlx5dv_hws_parser_node *mlx5dv_hws_parser_node_create(mlx5dv_hws_parser_graph *graph,
                                                               mlx5dv_hws_parser_node_cfg *node_cfg)

    cdef int mlx5dv_hws_parser_node_destroy(mlx5dv_hws_parser_node *node)

    cdef mlx5dv_hws_parser_arc *mlx5dv_hws_parser_arc_create(mlx5dv_hws_parser_graph *graph,
                                                             mlx5dv_hws_parser_arc_cfg *arc_cfg)

    cdef int mlx5dv_hws_parser_arc_destroy(mlx5dv_hws_parser_arc *arc)

    cdef mlx5dv_hws_parser_sampler *mlx5dv_hws_parser_sampler_create(mlx5dv_hws_parser_node *node,
                                                        mlx5dv_hws_parser_sampler_cfg *sampler_cfg)

    cdef int mlx5dv_hws_parser_sampler_query(mlx5dv_hws_parser_sampler *sampler,
                                             e.mlx5dv_hws_field_name *parser_fname,
                                             e.mlx5dv_hws_field_name *parser_ok_bit_fname,
                                             e.mlx5dv_hws_field_header *field_header,
                                             uint32_t *modify_header_field)

    cdef int mlx5dv_hws_parser_sampler_destroy(mlx5dv_hws_parser_sampler *sampler)

    cdef int mlx5dv_hws_parser_graph_bind(mlx5dv_hws_parser_graph *graph)

    cdef int mlx5dv_hws_parser_graph_unbind(mlx5dv_hws_parser_graph *graph)

    cdef mlx5dv_hws_parser_sampler * \
    mlx5dv_hws_parser_geneve_tlv_options_create(mlx5dv_hws_context *ctx,
                                                mlx5dv_hws_parser_geneve_option_cfg *cfg)
